from typing import ClassVar, Optional
from pydantic import BaseModel, Field, constr
from openctp_ctp import tdapi


class CtpField(BaseModel):
    _ctp_type_: ClassVar[callable] = None

    def ctp_object(self) -> any:
        obj = self._ctp_type_()
        for key, value in self.dict().items():
            if value is not None:
                setattr(obj, key, value)
        return obj

    class Config:
        orm_mode = True


class ReqUserLoginField(CtpField):
    _ctp_type_ = tdapi.CThostFtdcReqUserLoginField
    
    TradingDay: Optional[constr(max_length=9)] = Field(None, description='交易日')
    BrokerID: Optional[constr(max_length=11)] = Field(None, description='经纪公司代码')
    UserID: Optional[constr(max_length=16)] = Field(None, description='用户代码')
    Password: Optional[constr(max_length=41)] = Field(None, description='密码')
    UserProductInfo: Optional[constr(max_length=11)] = Field(
        None, description='用户端产品信息'
    )
    InterfaceProductInfo: Optional[constr(max_length=11)] = Field(
        None, description='接口端产品信息'
    )
    ProtocolInfo: Optional[constr(max_length=11)] = Field(None, description='协议信息')
    MacAddress: Optional[constr(max_length=21)] = Field(None, description='Mac地址')
    OneTimePassword: Optional[constr(max_length=41)] = Field(None, description='动态密码')
    reserve1: Optional[constr(max_length=16)] = Field(None, description='保留的无效字段')
    LoginRemark: Optional[constr(max_length=36)] = Field(None, description='登录备注')
    ClientIPPort: Optional[int] = Field(None, description='终端IP端口')
    ClientIPAddress: Optional[constr(max_length=33)] = Field(None, description='终端IP地址')


class RspUserLoginField(CtpField):
    _ctp_type_ = tdapi.CThostFtdcRspUserLoginField
    
    TradingDay: Optional[constr(max_length=9)] = Field(None, description='交易日')
    LoginTime: Optional[constr(max_length=9)] = Field(None, description='登录成功时间')
    BrokerID: Optional[constr(max_length=11)] = Field(None, description='经纪公司代码')
    UserID: Optional[constr(max_length=16)] = Field(None, description='用户代码')
    SystemName: Optional[constr(max_length=41)] = Field(None, description='交易系统名称')
    FrontID: Optional[int] = Field(None, description='前置编号')
    SessionID: Optional[int] = Field(None, description='会话编号')
    MaxOrderRef: Optional[constr(max_length=13)] = Field(None, description='最大报单引用')
    SHFETime: Optional[constr(max_length=9)] = Field(None, description='上期所时间')
    DCETime: Optional[constr(max_length=9)] = Field(None, description='大商所时间')
    CZCETime: Optional[constr(max_length=9)] = Field(None, description='郑商所时间')
    FFEXTime: Optional[constr(max_length=9)] = Field(None, description='中金所时间')
    INETime: Optional[constr(max_length=9)] = Field(None, description='能源中心时间')
    SysVersion: Optional[constr(max_length=41)] = Field(None, description='后台版本信息')


class UserLogoutField(CtpField):
    _ctp_type_ = tdapi.CThostFtdcUserLogoutField
    
    BrokerID: Optional[constr(max_length=11)] = Field(None, description='经纪公司代码')
    UserID: Optional[constr(max_length=16)] = Field(None, description='用户代码')


class ReqAuthenticateField(CtpField):
    _ctp_type_ = tdapi.CThostFtdcReqAuthenticateField
    
    BrokerID: Optional[constr(max_length=11)] = Field(None, description='经纪公司代码')
    UserID: Optional[constr(max_length=16)] = Field(None, description='用户代码')
    UserProductInfo: Optional[constr(max_length=11)] = Field(
        None, description='用户端产品信息'
    )
    AuthCode: Optional[constr(max_length=17)] = Field(None, description='认证码')
    AppID: Optional[constr(max_length=33)] = Field(None, description='App代码')


class RspAuthenticateField(CtpField):
    _ctp_type_ = tdapi.CThostFtdcRspAuthenticateField
    
    BrokerID: Optional[constr(max_length=11)] = Field(None, description='经纪公司代码')
    UserID: Optional[constr(max_length=16)] = Field(None, description='用户代码')
    UserProductInfo: Optional[constr(max_length=11)] = Field(
        None, description='用户端产品信息'
    )
    AppID: Optional[constr(max_length=33)] = Field(None, description='App代码')
    AppType: Optional[constr(max_length=1)] = Field(None, description='App类型')


class AuthenticationInfoField(CtpField):
    _ctp_type_ = tdapi.CThostFtdcAuthenticationInfoField
    
    BrokerID: Optional[constr(max_length=11)] = Field(None, description='经纪公司代码')
    UserID: Optional[constr(max_length=16)] = Field(None, description='用户代码')
    UserProductInfo: Optional[constr(max_length=11)] = Field(
        None, description='用户端产品信息'
    )
    AuthInfo: Optional[constr(max_length=129)] = Field(None, description='认证信息')
    IsResult: Optional[int] = Field(None, description='是否为认证结果')
    AppID: Optional[constr(max_length=33)] = Field(None, description='App代码')
    AppType: Optional[constr(max_length=1)] = Field(None, description='App类型')
    reserve1: Optional[constr(max_length=16)] = Field(None, description='保留的无效字段')
    ClientIPAddress: Optional[constr(max_length=33)] = Field(None, description='终端IP地址')


class RspInfoField(CtpField):
    _ctp_type_ = tdapi.CThostFtdcRspInfoField
    
    ErrorID: Optional[int] = Field(None, description='错误代码')
    ErrorMsg: Optional[constr(max_length=81)] = Field(None, description='错误信息')
    
class ExchangeField(CtpField):
    _ctp_type_ = tdapi.CThostFtdcExchangeField
    
    ExchangeID: Optional[constr(max_length=9)] = Field(None, description='交易所代码')
    ExchangeName: Optional[constr(max_length=61)] = Field(None, description='交易所名称')
    ExchangeProperty: Optional[constr(max_length=1)] = Field(None, description='交易所属性')


class ProductField(CtpField):
    _ctp_type_ = tdapi.CThostFtdcProductField
    
    reserve1: Optional[constr(max_length=31)] = Field(None, description='保留的无效字段')
    ProductName: Optional[constr(max_length=21)] = Field(None, description='产品名称')
    ExchangeID: Optional[constr(max_length=9)] = Field(None, description='交易所代码')
    ProductClass: Optional[constr(max_length=1)] = Field(None, description='产品类型')
    VolumeMultiple: Optional[int] = Field(None, description='合约数量乘数')
    PriceTick: Optional[float] = Field(None, description='最小变动价位')
    MaxMarketOrderVolume: Optional[int] = Field(None, description='市价单最大下单量')
    MinMarketOrderVolume: Optional[int] = Field(None, description='市价单最小下单量')
    MaxLimitOrderVolume: Optional[int] = Field(None, description='限价单最大下单量')
    MinLimitOrderVolume: Optional[int] = Field(None, description='限价单最小下单量')
    PositionType: Optional[constr(max_length=1)] = Field(None, description='持仓类型')
    PositionDateType: Optional[constr(max_length=1)] = Field(None, description='持仓日期类型')
    CloseDealType: Optional[constr(max_length=1)] = Field(None, description='平仓处理类型')
    TradeCurrencyID: Optional[constr(max_length=4)] = Field(None, description='交易币种类型')
    MortgageFundUseRange: Optional[constr(max_length=1)] = Field(
        None, description='质押资金可用范围'
    )
    reserve2: Optional[constr(max_length=31)] = Field(None, description='保留的无效字段')
    UnderlyingMultiple: Optional[float] = Field(None, description='合约基础商品乘数')
    ProductID: Optional[constr(max_length=81)] = Field(None, description='产品代码')
    ExchangeProductID: Optional[constr(max_length=81)] = Field(
        None, description='交易所产品代码'
    )
    OpenLimitControlLevel: Optional[constr(max_length=1)] = Field(
        None, description='开仓量限制粒度'
    )
    OrderFreqControlLevel: Optional[constr(max_length=1)] = Field(
        None, description='报单频率控制粒度'
    )


class InstrumentField(CtpField):
    _ctp_type_ = tdapi.CThostFtdcInstrumentField
    
    reserve1: Optional[constr(max_length=31)] = Field(None, description='保留的无效字段')
    ExchangeID: Optional[constr(max_length=9)] = Field(None, description='交易所代码')
    InstrumentName: Optional[constr(max_length=21)] = Field(None, description='合约名称')
    reserve2: Optional[constr(max_length=31)] = Field(None, description='保留的无效字段')
    reserve3: Optional[constr(max_length=31)] = Field(None, description='保留的无效字段')
    ProductClass: Optional[constr(max_length=1)] = Field(None, description='产品类型')
    DeliveryYear: Optional[int] = Field(None, description='交割年份')
    DeliveryMonth: Optional[int] = Field(None, description='交割月')
    MaxMarketOrderVolume: Optional[int] = Field(None, description='市价单最大下单量')
    MinMarketOrderVolume: Optional[int] = Field(None, description='市价单最小下单量')
    MaxLimitOrderVolume: Optional[int] = Field(None, description='限价单最大下单量')
    MinLimitOrderVolume: Optional[int] = Field(None, description='限价单最小下单量')
    VolumeMultiple: Optional[int] = Field(None, description='合约数量乘数')
    PriceTick: Optional[float] = Field(None, description='最小变动价位')
    CreateDate: Optional[constr(max_length=9)] = Field(None, description='创建日')
    OpenDate: Optional[constr(max_length=9)] = Field(None, description='上市日')
    ExpireDate: Optional[constr(max_length=9)] = Field(None, description='到期日')
    StartDelivDate: Optional[constr(max_length=9)] = Field(None, description='开始交割日')
    EndDelivDate: Optional[constr(max_length=9)] = Field(None, description='结束交割日')
    InstLifePhase: Optional[constr(max_length=1)] = Field(None, description='合约生命周期状态')
    IsTrading: Optional[int] = Field(None, description='当前是否交易')
    PositionType: Optional[constr(max_length=1)] = Field(None, description='持仓类型')
    PositionDateType: Optional[constr(max_length=1)] = Field(None, description='持仓日期类型')
    LongMarginRatio: Optional[float] = Field(None, description='多头保证金率')
    ShortMarginRatio: Optional[float] = Field(None, description='空头保证金率')
    MaxMarginSideAlgorithm: Optional[constr(max_length=1)] = Field(
        None, description='是否使用大额单边保证金算法'
    )
    reserve4: Optional[constr(max_length=31)] = Field(None, description='保留的无效字段')
    StrikePrice: Optional[float] = Field(None, description='执行价')
    OptionsType: Optional[constr(max_length=1)] = Field(None, description='期权类型')
    UnderlyingMultiple: Optional[float] = Field(None, description='合约基础商品乘数')
    CombinationType: Optional[constr(max_length=1)] = Field(None, description='组合类型')
    InstrumentID: Optional[constr(max_length=81)] = Field(None, description='合约代码')
    ExchangeInstID: Optional[constr(max_length=81)] = Field(
        None, description='合约在交易所的代码'
    )
    ProductID: Optional[constr(max_length=81)] = Field(None, description='产品代码')
    UnderlyingInstrID: Optional[constr(max_length=81)] = Field(
        None, description='基础商品代码'
    )


class SpecificInstrumentField(CtpField):
    _ctp_type_ = tdapi.CThostFtdcSpecificInstrumentField
    
    reserve1: Optional[constr(max_length=31)] = Field(None, description='保留的无效字段')
    InstrumentID: Optional[constr(max_length=81)] = Field(None, description='合约代码')
    
    
class DepthMarketDataField(CtpField):
    _ctp_type_ = tdapi.CThostFtdcDepthMarketDataField
    
    TradingDay: Optional[constr(max_length=9)] = Field(None, description='交易日')
    reserve1: Optional[constr(max_length=31)] = Field(None, description='保留的无效字段')
    ExchangeID: Optional[constr(max_length=9)] = Field(None, description='交易所代码')
    reserve2: Optional[constr(max_length=31)] = Field(None, description='保留的无效字段')
    LastPrice: Optional[float] = Field(None, description='最新价')
    PreSettlementPrice: Optional[float] = Field(None, description='上次结算价')
    PreClosePrice: Optional[float] = Field(None, description='昨收盘')
    PreOpenInterest: Optional[float] = Field(None, description='昨持仓量')
    OpenPrice: Optional[float] = Field(None, description='今开盘')
    HighestPrice: Optional[float] = Field(None, description='最高价')
    LowestPrice: Optional[float] = Field(None, description='最低价')
    Volume: Optional[int] = Field(None, description='数量')
    Turnover: Optional[float] = Field(None, description='成交金额')
    OpenInterest: Optional[float] = Field(None, description='持仓量')
    ClosePrice: Optional[float] = Field(None, description='今收盘')
    SettlementPrice: Optional[float] = Field(None, description='本次结算价')
    UpperLimitPrice: Optional[float] = Field(None, description='涨停板价')
    LowerLimitPrice: Optional[float] = Field(None, description='跌停板价')
    PreDelta: Optional[float] = Field(None, description='昨虚实度')
    CurrDelta: Optional[float] = Field(None, description='今虚实度')
    UpdateTime: Optional[constr(max_length=9)] = Field(None, description='最后修改时间')
    UpdateMillisec: Optional[int] = Field(None, description='最后修改毫秒')
    BidPrice1: Optional[float] = Field(None, description='申买价一')
    BidVolume1: Optional[int] = Field(None, description='申买量一')
    AskPrice1: Optional[float] = Field(None, description='申卖价一')
    AskVolume1: Optional[int] = Field(None, description='申卖量一')
    BidPrice2: Optional[float] = Field(None, description='申买价二')
    BidVolume2: Optional[int] = Field(None, description='申买量二')
    AskPrice2: Optional[float] = Field(None, description='申卖价二')
    AskVolume2: Optional[int] = Field(None, description='申卖量二')
    BidPrice3: Optional[float] = Field(None, description='申买价三')
    BidVolume3: Optional[int] = Field(None, description='申买量三')
    AskPrice3: Optional[float] = Field(None, description='申卖价三')
    AskVolume3: Optional[int] = Field(None, description='申卖量三')
    BidPrice4: Optional[float] = Field(None, description='申买价四')
    BidVolume4: Optional[int] = Field(None, description='申买量四')
    AskPrice4: Optional[float] = Field(None, description='申卖价四')
    AskVolume4: Optional[int] = Field(None, description='申卖量四')
    BidPrice5: Optional[float] = Field(None, description='申买价五')
    BidVolume5: Optional[int] = Field(None, description='申买量五')
    AskPrice5: Optional[float] = Field(None, description='申卖价五')
    AskVolume5: Optional[int] = Field(None, description='申卖量五')
    AveragePrice: Optional[float] = Field(None, description='当日均价')
    ActionDay: Optional[constr(max_length=9)] = Field(None, description='业务日期')
    InstrumentID: Optional[constr(max_length=81)] = Field(None, description='合约代码')
    ExchangeInstID: Optional[constr(max_length=81)] = Field(
        None, description='合约在交易所的代码'
    )
    BandingUpperPrice: Optional[float] = Field(None, description='上带价')
    BandingLowerPrice: Optional[float] = Field(None, description='下带价')



from pydantic import BaseModel
from .enums import CtpMethod
from .fields import *


class CtpResponse(BaseModel):
    method: CtpMethod
    
    RspInfo: RspInfoField = RspInfoField()
    RequestID: Optional[int] = None
    IsLast: bool = True
    
    @property
    def args(self) -> list[any]:
        return []


class RspUserLogin(CtpResponse):
    method: CtpMethod = CtpMethod.OnRspUserLogin

    RspUserLogin: Optional[RspUserLoginField]
    
    @property
    def args(self) -> list[any]:
        return [self.RspUserLogin, self.RspInfo, self.RequestID, self.IsLast]


class RspSubMarketData(CtpResponse):
    method: CtpMethod = CtpMethod.OnRspSubMarketData
    
    SpecificInstrument: Optional[SpecificInstrumentField]
    
    @property
    def args(self) -> list[any]:
        return [self.SpecificInstrument, self.RspInfo, self.RequestID, self.IsLast]


class RtnDepthMarketData(CtpResponse):
    method: CtpMethod = CtpMethod.OnRtnDepthMarketData
    
    DepthMarketData: Optional[DepthMarketDataField]
    
    @property
    def args(self) -> list[any]:
        return [self.DepthMarketData]



from openctp_client.objects import *
from openctp_client.objects.enums import CtpMethod
from openctp_client.clients.simple_ctp_client import MdClient

if __name__ == "__main__":
    config = CtpConfig("tcp://180.168.146.187:10211", "9999", "9999", "9999", "9999", "9999")
    client = MdClient(config)
    
    def connected(login_info: RspUserLoginField, rsp_info: RspInfoField, request_id: int, last: bool):
        print(f"connected: {login_info.TradingDay}")
        client.SubscribeMarketData(["ag2308"])
    
    def on_subscribe_market_data(instrument: SpecificInstrumentField, rsp_info: RspInfoField, request_id: int, last: bool):
        print(f"subscribe market data: {instrument.InstrumentID}")
    
    def on_market_data(data: DepthMarketDataField):
        print(f"market data: {data.InstrumentID} {data.AskPrice1}")
    
    client.add_spi_callback(CtpMethod.OnRspUserLogin, connected)
    client.add_spi_callback(CtpMethod.OnRspSubMarketData, on_subscribe_market_data)
    client.add_spi_callback(CtpMethod.OnRtnDepthMarketData, on_market_data)
    
    client.Connect()
    
    ch = input("press any key to exit")



from enum import Enum, auto


class CtpMethod(Enum):
    Connect = auto()
    Disconnect = auto()
    SubscribeMarketData = auto()
    OnRspUserLogin = auto()
    OnRspSubMarketData = auto()
    OnRtnDepthMarketData = auto()

    ReqSettlementInfoConfirm = auto()
    OnRspSettlementInfoConfirm = auto()
    ReqOrderInsert = auto()
    OnOrderInsert = auto()
    OnErrRtnOrderInsert = auto()
    OnRtnTrade = auto()
    
    @classmethod
    def nameOf(cls, name):
        return cls.__members__.get(name, None)



from pydantic import BaseModel
from .enums import CtpMethod
from .fields import *


class CtpRequest(BaseModel):
    method: CtpMethod
    
    @property
    def args(self) -> list[any]:
        return []



from typing import Any, Callable, Optional, Tuple
from openctp_ctp import mdapi

from ..objects.config import CtpConfig
from ..objects.enums import CtpMethod
from ..objects.fields import *
from ..objects.responses import *


class MdClient(mdapi.CThostFtdcMdSpi):
    
    def __init__(self, config: CtpConfig) -> None:
        super().__init__()
        self.config: CtpConfig = config
        self.request_count: int = 0
        self._callback: Callable[[CtpResponse], None] = self._default_callback
        self._call_map: dict[CtpMethod, list[Callable]] = {}
        self.api: mdapi.CThostFtdcMdApi = mdapi.CThostFtdcMdApi.CreateFtdcMdApi(self.config.user_id)
        self.api.RegisterSpi(self)
        self.api.RegisterFront(self.config.addr)
    
    @property
    def request_id(self) -> int:
        self.request_count += 1
        return self.request_count
    
    @property
    def callback(self) -> Callable[[CtpResponse], None]:
        return self._callback
    
    @callback.setter
    def callback(self, callback: Callable[[CtpResponse], None]) -> None:
        self._callback = callback
    
    def _default_callback(self, response: CtpResponse) -> None:
        if response.method in self._call_map:
            for callback in self._call_map[response.method]:
                callback(*response.args)
        else:
            # TODO: add warning
            print(f"no callback for {response.method.name()} found")
    
    def add_spi_callback(self, method: CtpMethod, callback: Callable) -> None:
        if method in self._call_map:
            self._call_map[method].append(callback)
        else:
            self._call_map[method] = [callback]
    
    def get_spi_callback(self, method: CtpMethod) -> Optional[list[Callable]]:
        return self._call_map.get(method, None)
    
    def del_spi_callback(self, method: CtpMethod, callback: Callable):
        if method in self._call_map:
            self._call_map[method].remove(callback)
            
    def Connect(self) -> None:
        self.api.Init()
    
    def Disconnect(self) -> None:
        self.api.Release()
        self.api.Join()
    
    def OnFrontConnected(self):
        self._login()

    def OnFrontDisconnected(self, nReason):
        # TODO: add log
        pass
    
    def _login(self):
        req = mdapi.CThostFtdcReqUserLoginField()
        req.BrokerID = self.config.broker_id
        req.UserID = self.config.user_id
        req.Password = self.config.password
        self.api.ReqUserLogin(req, self.request_id)

    def OnRspUserLogin(self, pRspUserLogin: mdapi.CThostFtdcRspUserLoginField, pRspInfo: mdapi.CThostFtdcRspInfoField, nRequestID: int, bIsLast: bool):
        """called when login responding"""
        rsp = RspUserLogin(RequestID=nRequestID, IsLast=bIsLast)
        if pRspUserLogin:
            rsp.RspUserLogin = RspUserLoginField.from_orm(pRspUserLogin)
            
        if pRspInfo is not None:
            rsp.RspInfo = RspInfoField.from_orm(pRspInfo)
            print(f"login rsp info, ErrorID: {pRspInfo.ErrorID}, ErrorMsg: {pRspInfo.ErrorMsg}")
            
        self.callback(rsp)
   
    def SubscribeMarketData(self, instrument_ids: list[str]) -> Tuple[int, int]:
        instrument_ids = list(map(lambda i: i.encode(), instrument_ids))
        request_id = self.request_id
        ret = self.api.SubscribeMarketData(instrument_ids, request_id)
        return (request_id, ret)
    
    def OnRspSubMarketData(self, pSpecificInstrument: mdapi.CThostFtdcSpecificInstrumentField, pRspInfo: mdapi.CThostFtdcRspInfoField, nRequestID, bIsLast):
        rsp = RspSubMarketData(RequestID=nRequestID, IsLast=bIsLast)
        if pSpecificInstrument:
            rsp.SpecificInstrument = SpecificInstrumentField.from_orm(pSpecificInstrument)
            
        if pRspInfo:
            rsp.RspInfo = RspInfoField.from_orm(pRspInfo)
        
        self.callback(rsp)

    def OnRtnDepthMarketData(self, pDepthMarketData: mdapi.CThostFtdcDepthMarketDataField):
        rsp = RtnDepthMarketData()
        rsp.DepthMarketData = DepthMarketDataField.from_orm(pDepthMarketData)
        self.callback(rsp)

            
    # 理论上用户回调和SPI回调函数重名没有问题，但是这样真的好吗？
    # def __setattr__(self, __name: str, __value: Any) -> None:
    #     method = CtpMethod.nameOf(__name)
    #     if method:
    #         self._call_map[method] = __value
    #     else:
    #         return super().__setattr__(__name, __value)
    
    # def __getattribute__(self, __name: str) -> Any:
    #     method = CtpMethod.nameOf(__name)
    #     if method and method in self._call_map:
    #         return self._call_map[method]
    #     else:
    #         return super().__getattribute__(__name)
    
    # def __delattr__(self, __name: str) -> None:
    #     method = CtpMethod.nameOf(__name)
    #     if method and method in self._call_map:
    #         del self._call_map[method]
    #     else:
    #         return super().__delattr__(__name)



class CtpConfig(object):

    def __init__(self, addr, broker_id, auth_code, app_id, user_id, password) -> None:
        self.addr = addr
        self.broker_id = broker_id
        self.auth_code = auth_code
        self.app_id = app_id
        self.user_id = user_id
        self.password = password
